-- To do:
--   o  Replace use of Pointer allocation with access all into a converted
--      input string.
--   o  Put parsing code into a separate private child package.
--   o  Is My_URL the best place (and name) for this utility?
--   o  Make Mosaic compatible

with Ada.Strings.Fixed;
with CGI.Characters;

package body CGI.Parameters is

  type UString is access String; -- Never deallocated!
  function To_String (U : UString) return String is
  begin
    if U = null then return "";
    else             return U.all;
    end if;
  end;

  subtype Table_Index is Natural; 
  type Key_Value_Pair is
    record
      Key        : UString;
      Value      : UString;
      Next_Key   : Table_Index := 0;
      Next_Value : Table_Index := 0;
    end record;

  --------------------------
  -- The following are set once during parsing. Each has a corresponding
  -- function in the specification to return the value.
  Parsing_Errors_Occurred : Boolean         := False;
  Input_Was_Received      : Boolean         := False;
  Is_Index_Request_Made   : Boolean         := False;

  type Key_Value_Table is array (Positive range <>) of Key_Value_Pair;
  Empty_Table : Key_Value_Table(1..0);
  
  ------------------------------------------------------------------
  -- This package automatically parses CGI input on program start.
  function Parse return Key_Value_Table is

    function Field_End(Data           : String;
                       Field_Separator: Character) return Natural is
    -- Return the position of first Field_Separator in Data.
    -- If there's no Field_Separator, return Data'Last +1.
    begin
      for Pos in Data'Range loop
        if Data(Pos) = Field_Separator then
           return Pos;
        end if;
      end loop;
      return Data'Last +1;
    end Field_End;
    
    Last_Key_Added : Table_Index := 0; -- Use this to set Next_Key
    procedure Set_CGI_Position(CGI_Data   : in out Key_Value_Table;
                               Key_Number : in     Positive;
                               Datum      : in     String) is
      -- Given a Key number and a datum of the form key=value
      -- assign the CGI_Data(Key_Number) the value of key and value.
      Pos  : constant Natural := Field_End(Datum, '=');
      LKey : constant String
           := CGI.Characters.Decode(Datum(Datum'First..Pos-1));
      Value: constant String
           := CGI.Characters.Decode(Datum(Pos+1..Datum'Last));

      function Find_Previous_Key return Table_Index is -- optimization code
        Position : Table_Index := Key_Number;
      begin
        if Position = 1 then return 0;
        elsif To_String(CGI_Data(Position-1).Key) = LKey then
          -- the likely and efficient has occured
          return Position -1;
        else
          Position := 1;
          Find_First_Key_Position:
          while Position /= 0 loop
            exit when To_String(CGI_Data(Position).Key) = LKey;
            Position := CGI_Data(Position).Next_Key;
          end loop Find_First_Key_Position;

          if Position = 0 then -- the most  likely but more difficult to detect
            return 0;
          else                 -- the least likely and more difficult to detect
            while CGI_Data(Position).Next_Value /= 0 loop
              Position := CGI_Data(Position).Next_Value;
            end loop;
            return Position;
          end if;
        end if;
      end Find_Previous_Key;
    begin
      Set_Key:
      declare
        Position : constant Table_Index := Find_Previous_Key;
      begin
        if Position /= 0 then -- This isn't the first occurence of this key
          CGI_Data(Key_Number).Key := CGI_Data(Position).Key; -- reuse the pointer
          CGI_Data(Position).Next_Value := Key_Number;        -- Link'em!
        else                  -- This is the first occurence of this key
          CGI_Data(Key_Number).Key := new String'(LKey);
          if Last_Key_Added /= 0 then -- This isn't the first Key though
            CGI_Data(Last_Key_Added).Next_Key := Key_Number;
          end if;
          Last_Key_Added := Key_Number;
        end if;
      end Set_Key;

      CGI_Data(Key_Number).Value := new String'(Value);
    end Set_CGI_Position;
    

    function Create_CGI_Data (Input : in  String) return Key_Value_Table is
      CGI_Data : Key_Value_Table(1..Ada.Strings.Fixed.Count(Input, "&")+1);
      Pos      : Positive := Input'First;
      Last     : Natural;
    begin
      for Key_Number in CGI_Data'range loop
        exit when Pos > Input'Last;

        Last := Field_End(Input(Pos..Input'Last), '&');
        Set_CGI_Position (CGI_Data, Key_Number, Input(Pos..Last-1));
        Pos := Last + 1; -- Skip over field separator.
      end loop;
      return CGI_Data;
    end Create_CGI_Data;

  begin -- Parse
    if Input = "" then
      return Empty_Table;
    else
      Input_Was_Received := True;
      if Ada.Strings.Fixed.Index(Input, "=") = 0 then
        -- No "=" found, so this is an "Isindex" request.
        -- An "Isindex" request is turned into a Key of "isindex" at position 1,
        -- with Value(1) as the actual query.???
        Is_Index_Request_Made := True;
        return Create_CGI_Data("isindex=" & Input);
      else
        return Create_CGI_Data(Input);
      end if;

    end if;

  exception
    when others =>
      Parsing_Errors_Occurred := True;
      return Empty_Table;
  end Parse;


  CGI_Data : Key_Value_Table := Parse;
  --------------------------------------------------------
  function Parsing_Errors return Boolean is
  begin
     return Parsing_Errors_Occurred;
  end Parsing_Errors;
  
  function Input_Received return Boolean is
  begin
    return Input_Was_Received;
  end Input_Received;
  
  
  function Is_Index return Boolean is
  begin
    return Is_Index_Request_Made;
  end Is_Index;


  -- This utility is very useful for implementing the value selector operations
  function Find_Position (Key   : in String;
                          Index : in Positive) return Table_Index is
    Position : Table_Index := 0;
  begin
    if CGI_Data /= Empty_Table then
      Position := 1;
      Find_First_Key_Position:
      while Position /= 0 loop
        exit when To_String(CGI_Data(Position).Key) = Key;
        Position := CGI_Data(Position).Next_Key;
      end loop Find_First_Key_Position;
  
      for Count in 2..Index loop
         exit when Position = 0;
         Position := CGI_Data(Position).Next_Value;
      end loop;
    end if;

    return Position;  -- 0 if not found  
  end Find_Position;

  function Value(Key   : in String;
                 Index : in Positive := 1) return String is
    Key_Position : constant Table_Index := Find_Position (Key, Index);
  begin
    if Key_Position = 0 then raise  Constraint_Error;
    else                     return To_String(CGI_Data(Key_Position).Value);
    end if;

  end Value;
  
  
  function Key_Exists(Key   : in String;
                      Index : in Positive := 1) return Boolean is
  begin
    return Find_Position(Key,Index) /= 0;
  end Key_Exists;
  
  
  function Value_Count(Key : in String) return Natural is
    Count    : Natural     := 0;
    Position : Table_Index := Find_Position(Key, 1);
  begin

    while Position /= 0 loop
      Count := Count +1;
      Position := CGI_Data(Position).Next_Value;
    end loop;

    return Count;
  end Value_Count;
  
  procedure Iterate_Key (Key : in String) is
    Position : Table_Index := Find_Position(Key, 1);
  begin
    while Position /= 0 loop
      Evaluate (To_String(CGI_Data(Position).Value));
      Position := CGI_Data(Position).Next_Value;
    end loop;
  end Iterate_Key;
  
  
  -----------------------------------------------------------
  function Argument_Count return Natural is
  begin
    return CGI_Data'Length;
  end Argument_Count;
  
  
  function Key(Position : in Positive) return String is
  begin
    return To_String(CGI_Data(Position).Key);
  end Key;
  
  
  function Value(Position : in Positive) return String is
  begin
    return To_String(CGI_Data(Position).Value);
  end Value;
  
  procedure Iterate is
  begin
    for Position in CGI_Data'range loop
      Evaluate (To_String(CGI_Data(Position).Key),
                To_String(CGI_Data(Position).Value));
    end loop;
  end Iterate;

end CGI.Parameters;
