-- get information from CGI interface to create a command to
-- compile an Ada file. The command is written to Standard
-- Output. Errors use echo to describe any problems with
-- parameters. The syntax for Compile is:
--   compile?(file=<URL>)|code=<text>)[&options=<text>]
--   [&formatter=<URL>]
--   [&head=<text>][&foot=<text>]
--
-- The command will:
--   The head text is written first.
--   When code is provided, gnatchop splits it up into files for compiling,
--   then the file(s) are compiled using the options.
--   The output is processed by the formatter and written second.
--   Finally, the foot text is written.

with CGI.Parameters;
with CGI.Output;
with HTML;
with Ada.Text_IO;
with Ada.Exceptions;
with GNAT_Compiler;

procedure Compile is

   Temporary_File   : constant String := "temp"; -- No extension allowed!
   Source_Directory : constant String := "/export/user_data/ada_source/";
   
   -- CGI Parameter names(multiple parameters are handled intuitively):
   File      : constant String := "file";    -- Each file is compiled
   Code      : constant String := "code";    -- Saved in the first file or else
                                             --   the temporary file
   Option    : constant String := "option";  -- All added to -gnat
   Head      : constant String := "head";    -- Echo'd after mime
   Error     : constant String := "error";   -- Echo'd after errors in compile
   Success   : constant String := "success"; -- Echo'd after successful compile
   Foot      : constant String := "foot";    -- Echo'd last

   procedure Echo (Message : in String) is
   begin
      CGI.Output.Put_Line ( HTML.Paragraph & HTML.Bold(Message));
   end;

   procedure Echo_Value is
     new CGI.Parameters.Iterate_Key (CGI.Output.Put_Line);

   function Extract_Options return String is
      Count : constant Natural := CGI.Parameters.Value_Count (Option);
      function Append (Index : in Positive) return String is
      begin
         if Index <= Count then
            return CGI.Parameters.Value (Option,Index) & Append (Index +1);
         else
            return "";
         end if;
      end;
   begin
      return Append(1);
   end Extract_Options;

   procedure Compile_Source (File_Name : in String) is
      Has_Errors : Boolean;
   begin
      GNAT_Compiler.Compile (Source_Directory & File_Name,
                             Extract_Options,
                             Has_Errors);
      if Has_Errors then
         Echo_Value(Error);
      else
         Echo_Value(Success);
      end if;
   end;

   procedure Compile_Files is
      new CGI.Parameters.Iterate_Key (Compile_Source);
      
   function Name_to_Save_Code return String is
   begin
      if CGI.Parameters.Key_Exists (File) then -- use it to save the code
         return CGI.Parameters.Value (File);
      else
         return Temporary_File;
      end if;
   end Name_to_Save_Code;

   procedure Save_Code is
      use Ada.Text_IO;
      Output_Name : constant String := Source_Directory & Name_to_Save_Code;
      Output_File : File_Type;

      procedure Copy_to_File (Value : in String) is
      begin
         Put_Line (Output_File, Value);
      end;
      procedure Copy_Every_Value is
         new CGI.Parameters.Iterate_Key (Copy_to_File);
   begin
      Create (Output_File, Out_File, Output_Name);
      Copy_Every_Value (Code);
      Close (Output_File);

      Compile_Source (Name_to_Save_Code);

   exception
      when Error : others =>
         echo (Ada.Exceptions.Exception_Name(Error) &
               " while saving code to file: " & Output_Name);
   end Save_Code;

  use CGI.Parameters;
begin
  CGI.Output.Content_Text_HTML;
  Echo_Value (Head);
   
  if not Input_Received           then echo ("No input received");
  elsif Parsing_Errors            then echo ("Parsing Errors");
  elsif not Key_Exists (File) and
        not Key_Exists (Code)     then echo ("Nothing to compile");
  else

    if Key_Exists(Code) then
      Save_Code;
    else
      Compile_Files (File);
    end if;

  end if;

  Echo_Value (Foot);

exception
  when Unexpected_Exception : others =>
    echo (Ada.Exceptions.Exception_Name(Unexpected_Exception) &
          " exception in Compile.");
    Echo_Value (Foot);
end Compile;
